// I2C device class (I2Cdev) demonstration Arduino sketch for MPU6050 class
// 10/7/2011 by Jeff Rowberg <jeff@rowberg.net>
// Updates should (hopefully) always be available at https://github.com/jrowberg/i2cdevlib

// I2Cdev and MPU6050 must be installed as libraries, or else the .cpp/.h files
// for both classes must be in the include path of your project
#include "I2Cdev.h"
#include "MPU6050.h"

// Arduino Wire library is required if I2Cdev I2CDEV_ARDUINO_WIRE implementation
// is used in I2Cdev.h
#if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
    #include "Wire.h"
#endif

// class default I2C address is 0x68
// specific I2C addresses may be passed as a parameter here
// AD0 low = 0x68 (default for InvenSense evaluation board)
// AD0 high = 0x69
MPU6050 accelgyro;
//MPU6050 accelgyro(0x69); // <-- use for AD0 high

int16_t ax, ay, az;
int16_t gx, gy, gz;


#include <Servo.h> 
#define SERVO_CAM       4
#define SERVO_DRIVE_1   5
#define SERVO_DRIVE_2   6
#define SERVO_DRIVE_3   7
#define SERVO_DRIVE_4   8
#define RASPI_POWER     9
#define LIGHT           3
#define COM_PULSE      13 
#define COM_RECEIVE    12 
#define COM_SEND       10

Servo ServoCam;
int ServoCamIsOff;
int driveDirection = 0;
Servo ServoDrive1;
Servo ServoDrive2;
Servo ServoDrive3;
Servo ServoDrive4;

double gyroTurnX=0.0;
double gyroTurnY=90.0;
double gyroTurnZ=0.0;
double totalTilt=0.0;
int mapY;
byte comByte = 0;
int bitCount = 0;
int lastComState = 0;
double turnDeg = 0.0;
uint32_t mTimerCom;
uint32_t uTimer;
uint32_t uTimerNow;
uint32_t servoTimer;
uint32_t raspberryTimer;

void setup() {
    // join I2C bus (I2Cdev library doesn't do this automatically)
    #if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
        Wire.begin();
    #elif I2CDEV_IMPLEMENTATION == I2CDEV_BUILTIN_FASTWIRE
        Fastwire::setup(400, true);
    #endif

    // initialize serial communication
    Serial.begin(115200);

    // initialize device
    Serial.println("Initializing I2C devices...");
    accelgyro.initialize();

    //Raw data -32768, +32767
    //accelgyro.setFullScaleAccelRange(MPU6050_ACCEL_FS_8); // +/- 2, 4, 8, or 16g for the accelerometer
    accelgyro.setFullScaleGyroRange(MPU6050_GYRO_FS_2000); // +/- 250, 500, 1000, or 2000 deg/sec for the gyroscope.
    
    // verify connection
    Serial.println("Testing device connections...");
    Serial.println(accelgyro.testConnection() ? "MPU6050 connection successful" : "MPU6050 connection failed");

    lastComState = 0;
    bitCount = 0;
    comByte = 0;
    turnDeg = 0;
    gyroTurnX = 0.0;
    gyroTurnZ = 0.0;
    driveDirection = 0;
    
    servoTimer = millis();
    pinMode(LIGHT, OUTPUT);
    pinMode(RASPI_POWER, OUTPUT);
    pinMode(COM_PULSE, INPUT);
    digitalWrite(COM_PULSE, 0);
    pinMode(COM_RECEIVE, INPUT);
    digitalWrite(COM_RECEIVE, 0);
    pinMode(COM_SEND, INPUT);
    digitalWrite(COM_SEND, 0);
    analogWrite(LIGHT, 128);
    delay(500);
    analogWrite(LIGHT, 0);
    digitalWrite(RASPI_POWER, 0);
    gyroTurnY=0.0;
    ServoCam.attach(SERVO_CAM);
    ServoCam.write(90);
    ServoCamIsOff = 0;
    
}

void loop() {
    //Observe communication
    if(digitalRead(COM_PULSE) == 1 && lastComState == 0){
      if(millis() - mTimerCom > 1000 || mTimerCom == 0){
        bitCount = 0;
        comByte = 0;        
        mTimerCom = millis();
      }
      lastComState = 1;
      if(digitalRead(COM_RECEIVE) == 1){
        comByte += (1<<(bitCount));
      }
      bitCount++;
      if(bitCount == 8){//Byte complete
        raspberryTimer = millis();
        if(comByte == 1){
          analogWrite(LIGHT, 10);
        }
        if(comByte == 2){
          analogWrite(LIGHT, 50);
        }
        if(comByte == 3){
          analogWrite(LIGHT, 100);
        }
        if(comByte == 4){
          analogWrite(LIGHT, 255);
        }
        if(comByte == 5){
          analogWrite(LIGHT, 0);
        }

        if(comByte > 5 && comByte < 16){//Drive forward
          driveDirection = 1;
          servoTimer = millis() + 500 * (comByte - 5);
          ServoDrive1.attach(SERVO_DRIVE_1);
          ServoDrive1.write(170);
          ServoDrive2.attach(SERVO_DRIVE_2);
          ServoDrive2.write(10);
          ServoDrive3.attach(SERVO_DRIVE_3);
          ServoDrive3.write(10);
          ServoDrive4.attach(SERVO_DRIVE_4);
          ServoDrive4.write(170);
        }
        if(comByte > 15 && comByte < 26){//Drive backward
          driveDirection = 2;
          servoTimer = millis() + 500 * (comByte - 15);
          ServoDrive1.attach(SERVO_DRIVE_1);
          ServoDrive1.write(10);
          ServoDrive2.attach(SERVO_DRIVE_2);
          ServoDrive2.write(170);
          ServoDrive3.attach(SERVO_DRIVE_3);
          ServoDrive3.write(170);
          ServoDrive4.attach(SERVO_DRIVE_4);
          ServoDrive4.write(10);
        }

        if(comByte > 25 && comByte < 36){//Turn right
          servoTimer = millis() + 2000;//Turn servos off after 2 sec
          turnDeg = 10 * (comByte - 25);
          gyroTurnY = 0.0;
          ServoDrive1.attach(SERVO_DRIVE_1);
          ServoDrive1.write(170);
          ServoDrive2.attach(SERVO_DRIVE_2);
          ServoDrive2.write(170);
          ServoDrive3.attach(SERVO_DRIVE_3);
          ServoDrive3.write(170);
          ServoDrive4.attach(SERVO_DRIVE_4);
          ServoDrive4.write(170);
        }
        if(comByte > 35 && comByte < 46){//Turn left
          servoTimer = millis() + 2000;//Turn servos off after 2 sec
          turnDeg = -10 * (comByte - 35);
          gyroTurnY = 0.0;
          ServoDrive1.attach(SERVO_DRIVE_1);
          ServoDrive1.write(10);
          ServoDrive2.attach(SERVO_DRIVE_2);
          ServoDrive2.write(10);
          ServoDrive3.attach(SERVO_DRIVE_3);
          ServoDrive3.write(10);
          ServoDrive4.attach(SERVO_DRIVE_4);
          ServoDrive4.write(10);
        }

        if(comByte > 45 && comByte < 66){//Camera angle
          if(ServoCamIsOff == 1){
            ServoCamIsOff = 0;
            ServoCam.attach(SERVO_CAM);
          }
          ServoCam.write(60 + 4 * (comByte - 45));
        }
        if(comByte == 66){//Turn off camera servo
          ServoCam.detach();
          ServoCamIsOff = 1;
        }

        mTimerCom = 0;
        
      }
    }
    if(digitalRead(COM_PULSE) == 0){
      lastComState = 0;
    }
  
    // read raw accel/gyro measurements from device
    accelgyro.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);

    // these methods (and a few others) are also available
    //accelgyro.getAcceleration(&ax, &ay, &az);
    //accelgyro.getRotation(&gx, &gy, &gz);


//    gyroTurnY += (gy-Y_OFFSET)*M_PI/180;
    double gyroXrate = -((double)gx/16.0);
    double gyroYrate = -((double)gy/16.0);
    double gyroZrate = -((double)gz/16.0);
    uTimerNow = micros();
    if(uTimerNow > uTimer){
      gyroTurnX += gyroXrate*((double)(uTimerNow-uTimer)/1000000);
      gyroTurnY += gyroYrate*((double)(uTimerNow-uTimer)/1000000);
      gyroTurnZ += gyroZrate*((double)(uTimerNow-uTimer)/1000000);
//      mapY = map(gyroTurnY, 0, 180, 0, 179);
//      ServoCam.write(mapY);
    }
    uTimer = micros();

    if(turnDeg != 0.0){
      if(turnDeg > 0.0 && gyroTurnY >= turnDeg){
        ServoDrive1.detach();
        ServoDrive2.detach();
        ServoDrive3.detach();
        ServoDrive4.detach();
        turnDeg = 0.0;
      }
      if(turnDeg < 0.0 && gyroTurnY <= turnDeg){
        ServoDrive1.detach();
        ServoDrive2.detach();
        ServoDrive3.detach();
        ServoDrive4.detach();
        turnDeg = 0.0;
      }
    }
    
    if(millis() > servoTimer){
      gyroTurnX = 0.0;
      gyroTurnZ = 0.0;
      ServoDrive1.detach();
      ServoDrive2.detach();
      ServoDrive3.detach();
      ServoDrive4.detach();
    }
    else{
      totalTilt = abs(gyroTurnX) + abs(gyroTurnZ);
      if(totalTilt > 6.0){//Rover tilts

        if(driveDirection == 2){
          driveDirection = 0;
          ServoDrive1.detach();
          ServoDrive2.detach();
          ServoDrive3.detach();
          ServoDrive4.detach();
          delay(20);

          ServoDrive1.attach(SERVO_DRIVE_1);
          ServoDrive1.write(170);
          ServoDrive2.attach(SERVO_DRIVE_2);
          ServoDrive2.write(10);
          ServoDrive3.attach(SERVO_DRIVE_3);
          ServoDrive3.write(10);
          ServoDrive4.attach(SERVO_DRIVE_4);
          ServoDrive4.write(170);
          servoTimer = millis() + 700;
        }
        if(driveDirection == 1){
          driveDirection = 0;

          ServoDrive1.detach();
          ServoDrive2.detach();
          ServoDrive3.detach();
          ServoDrive4.detach();
          delay(20);

          ServoDrive1.attach(SERVO_DRIVE_1);
          ServoDrive1.write(10);
          ServoDrive2.attach(SERVO_DRIVE_2);
          ServoDrive2.write(170);
          ServoDrive3.attach(SERVO_DRIVE_3);
          ServoDrive3.write(170);
          ServoDrive4.attach(SERVO_DRIVE_4);
          ServoDrive4.write(10);
          servoTimer = millis() + 700;
        }        
      }
    }
    
    if(millis() - raspberryTimer > 300000L){
      //uncomment line to activate Arduino watchdog!
      //digitalWrite(RASPI_POWER, 1);
      delay(5000);
      digitalWrite(RASPI_POWER, 0);
      raspberryTimer = millis();
    }
    if(millis() < raspberryTimer){
      raspberryTimer = millis();    
    }
}
